@@ -4,6 +4,7 @@
 #include <stdio.h>
 #include <string.h>
 #include <stdlib.h>
+#include "fossil.h"
 
 #define WORDS_STRUCT 0
 #define WORD_COLUMN_UTF8 1
@@ -11,16 +12,24 @@
 #define WTYPE_COLUMN 3
 #define FREQ_COLUMN 4
 #define CHECK_COLUMN 5
-#define NR_COLUMNS 6
+#define FIX_COLUMN 6
+#define NR_COLUMNS 7
 
 #define MAX_CANDIDATES 1000
 #define NR_BUCKETS 32768
 
+#define CHECKED_ITEM_FILE "checked_words"
+
 struct words {
     char *index;
     char *wt;
     int freq;
     char *word;
+    /**/
+    int is_top;
+    int checked;
+    int fixed;
+    /**/
     struct words *next;
     struct words *next_chain;
 };
@@ -33,6 +42,7 @@ struct app_win {
     /**/
     GtkWidget *top;
     GtkWidget *next_button;
+    GtkWidget *save_button;
     GtkWidget *quit_button;
     /**/
     GtkWidget *text_entry;
@@ -57,56 +67,6 @@ struct app_win {
     GtkWidget *conf_win;
 };
 
-struct syusi_pat {
-    char *wt;
-    char *tail;
-} syusi_pat[] = {
-    {"B5", ""},
-    {"B5r", ""},
-    {"C5r", ""},
-    {"G5", ""},
-    {"G5r", ""},
-    {"K5", ""},
-    {"K5r", ""},
-    {"KS", ""},
-    {"KSr", ""},
-    {"KY", ""},
-    {"KYT", ""},
-    {"KYU", ""},
-    {"KYme", ""},
-    {"KYmi", ""},
-    {"KYmime", ""},
-    {"KYna", ""},
-    {"L5", ""},
-    {"M5", ""},
-    {"M5r", ""},
-    {"N5", ""},
-    {"N5r", ""},
-    {"R5", ""},
-    {"R5r", ""},
-    {"SX", ""},
-    {"T5", ""},
-    {"T5r", ""},
-    {"U5", ""},
-    {"U5r", ""},
-    {"W5", ""},
-    {"W5r", ""},
-    {"ZX", ""},
-    {NULL, NULL}
-};
-
-static char *
-syusi_tail(const char *wt)
-{
-    struct syusi_pat *p;
-    for (p = syusi_pat; p->wt; p++) {
-	if (!strcmp(p->wt, &wt[1])) {
-	    return p->tail;
-	}
-    }
-    return "";
-}
-
 static guchar *
 euc_to_utf8 (const guchar *str)
 {
@@ -198,6 +158,9 @@ add_word(struct app_win *win,
     *w = w_tmp;
     w->index = strdup(index);
     w->word = strdup(word);
+    w->checked = 0;
+    w->fixed = 0;
+    w->is_top = 0;
     /* link */
     w->next = win->words.next;
     win->words.next = w;
@@ -207,10 +170,11 @@ add_word(struct app_win *win,
     win->dict.bucket[b] = w;
 }
 
-static void
+static int
 load_words_fp(FILE *fp, struct app_win *win)
 {
     char buf[1024];
+    int nr = 0;
     while (fgets(buf, 1024, fp)) {
 	char *str = buf;
 	char *index, *tmp;
@@ -227,9 +191,11 @@ load_words_fp(FILE *fp, struct app_win *win)
 		}
 	    } else {
 		add_word(win, index, wt, tmp);
+		nr ++;
 	    }
 	}
     }
+    return nr;
 }
 
 static void
@@ -239,6 +205,9 @@ update_model(struct app_win *win)
     GtkTreeIter iter;
     int i;
     for (w = win->words.next; w; w = w->next) {
+	if (w->checked) {
+	    continue;
+	}
 	gtk_list_store_append(GTK_LIST_STORE(win->words_model), &iter);
 	gtk_list_store_set(GTK_LIST_STORE(win->words_model), &iter,
 			   WORDS_STRUCT, w,
@@ -247,6 +216,7 @@ update_model(struct app_win *win)
 			   WTYPE_COLUMN, w->wt,
 			   FREQ_COLUMN, w->freq,
 			   CHECK_COLUMN, FALSE,
+			   FIX_COLUMN, FALSE,
 			   -1);
     }
 }
@@ -255,15 +225,17 @@ static void
 load_words(struct app_win *win)
 {
     int i;
+    int nr = 0;
     for (i = 1; i < win->argc; i++) {
 	char *fn = win->argv[i];
 	FILE *fp = fopen(fn, "r");
 	if (!fp) {
 	    continue;
 	}
-	load_words_fp(fp, win);
+	nr += load_words_fp(fp, win);
 	fclose(fp);
     }
+    printf("%d words\n", nr);
 }
 
 static void
@@ -277,15 +249,51 @@ next_cb(GtkWidget *widget, struct app_win *win)
 {
     GtkTreeIter iter;
     GtkTreeSelection *selection;
+    GValue value = {0, };
+    gboolean checked;
+    struct words *w;
     selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(win->words_view));
     if (!selection ||
 	!gtk_tree_selection_get_selected(selection, NULL, &iter)) {
 	return ;
     }
+    gtk_list_store_set (GTK_LIST_STORE (win->words_model), &iter,
+			CHECK_COLUMN, 1, -1);
+    gtk_tree_model_get_value(win->words_model, &iter,
+			     WORDS_STRUCT, &value);
+    w = g_value_get_pointer(&value);
+    w->checked = 1;
+    g_value_unset(&value);
+    /**/
     gtk_tree_model_iter_next(win->words_model, &iter);
     gtk_tree_selection_select_iter(selection, &iter);
 }
 
+static void
+save_cb(GtkWidget *widget, struct app_win *win)
+{
+    struct words *w;
+    FILE *fp = fopen(CHECKED_ITEM_FILE, "w");
+    if (!fp) {
+	printf("failed to open %s.\n", CHECKED_ITEM_FILE);
+	return ;
+    }
+    for (w = win->words.next; w; w = w->next) {
+	if (w->checked) {
+	    char *fix = "ok";
+	    if (w->fixed) {
+		if (w->is_top) {
+		    fix = "down";
+		} else {
+		    fix = "up";
+		}
+	    }
+	    fprintf(fp, "%s %s %s %s\n", w->index, w->word, w->wt, fix);
+	}
+    }
+    fclose(fp);
+}
+
 static void
 build_check_buttons(struct app_win *win, GtkWidget *hbox)
 {
@@ -296,12 +304,17 @@ build_check_buttons(struct app_win *win, GtkWidget *hbox)
     win->next_button = gtk_button_new_with_label("next");
     gtk_container_add(GTK_CONTAINER(vbox), win->next_button);
     gtk_widget_show(win->next_button);
+    win->save_button = gtk_button_new_with_label("save");
+    gtk_container_add(GTK_CONTAINER(vbox), win->save_button);
+    gtk_widget_show(win->save_button);
     win->quit_button = gtk_button_new_with_label("quit");
     gtk_container_add(GTK_CONTAINER(vbox), win->quit_button);
     gtk_widget_show(win->quit_button);
     /**/
     g_signal_connect(G_OBJECT(win->next_button), "clicked",
 		     G_CALLBACK(next_cb), win);
+    g_signal_connect(G_OBJECT(win->save_button), "clicked",
+		     G_CALLBACK(save_cb), win);
     g_signal_connect(G_OBJECT(win->quit_button), "clicked",
 		     G_CALLBACK(quit), NULL);
     /**/
@@ -318,28 +331,63 @@ create_words_model()
 			       G_TYPE_STRING,
 			       G_TYPE_STRING,
 			       G_TYPE_UINT,
+			       G_TYPE_BOOLEAN,
 			       G_TYPE_BOOLEAN);
 
     return GTK_TREE_MODEL(model);
 }
 
 static void
-toggled(GtkCellRendererToggle *toggle,
-	gchar *path_str,
-	struct app_win *win)
+fix_toggled(GtkCellRendererToggle *toggle,
+	    gchar *path_str,
+	    struct app_win *win)
 {
     GtkTreeIter iter;
     GtkTreePath *path = gtk_tree_path_new_from_string (path_str);
     gboolean fixed;
+    struct words *w;
+    GValue value = {0, };
+
+    gtk_tree_model_get_iter (win->words_model, &iter, path);
+    gtk_tree_model_get (win->words_model, &iter, FIX_COLUMN, &fixed, -1);
+
+    fixed = !fixed;
+    /**/
+    gtk_tree_model_get_value(win->words_model, &iter,
+			     WORDS_STRUCT, &value);
+    w = g_value_get_pointer(&value);
+    w->fixed = fixed;
+    g_value_unset(&value);
+    /* set to model */
+    gtk_list_store_set (GTK_LIST_STORE (win->words_model), &iter,
+			FIX_COLUMN, fixed, -1);
+    gtk_tree_path_free (path);
+}
+
+static void
+check_toggled(GtkCellRendererToggle *toggle,
+	      gchar *path_str,
+	      struct app_win *win)
+{
+    GtkTreeIter iter;
+    GtkTreePath *path = gtk_tree_path_new_from_string (path_str);
+    gboolean fixed;
+    struct words *w;
+    GValue value = {0, };
 
     gtk_tree_model_get_iter (win->words_model, &iter, path);
     gtk_tree_model_get (win->words_model, &iter, CHECK_COLUMN, &fixed, -1);
 
     fixed = !fixed;
-
+    /**/
+    gtk_tree_model_get_value(win->words_model, &iter,
+			     WORDS_STRUCT, &value);
+    w = g_value_get_pointer(&value);
+    w->checked = fixed;
+    g_value_unset(&value);
+    /* set to model */
     gtk_list_store_set (GTK_LIST_STORE (win->words_model), &iter,
 			CHECK_COLUMN, fixed, -1);
-
     gtk_tree_path_free (path);
 
 }
@@ -395,13 +443,27 @@ add_columns(struct app_win *win, GtkTreeView *treeview)
     gtk_tree_view_append_column (treeview, column);
     /**/
     renderer = gtk_cell_renderer_toggle_new();
+    column = gtk_tree_view_column_new_with_attributes ("fix!",
+						       renderer,
+						       "active",
+						       FIX_COLUMN,
+						       NULL);
+    g_signal_connect(renderer, "toggled",
+		     G_CALLBACK(fix_toggled), win);
+    gtk_tree_view_column_set_sizing (GTK_TREE_VIEW_COLUMN (column),
+				     GTK_TREE_VIEW_COLUMN_FIXED);
+    gtk_tree_view_column_set_fixed_width (GTK_TREE_VIEW_COLUMN (column), 50);
+    gtk_tree_view_column_set_sort_column_id(column, CHECK_COLUMN);
+    gtk_tree_view_append_column (treeview, column);
+    /**/
+    renderer = gtk_cell_renderer_toggle_new();
     column = gtk_tree_view_column_new_with_attributes ("Checked?",
 						       renderer,
 						       "active",
 						       CHECK_COLUMN,
 						       NULL);
     g_signal_connect(renderer, "toggled",
-		     G_CALLBACK(toggled), win);
+		     G_CALLBACK(check_toggled), win);
     gtk_tree_view_column_set_sizing (GTK_TREE_VIEW_COLUMN (column),
 				     GTK_TREE_VIEW_COLUMN_FIXED);
     gtk_tree_view_column_set_fixed_width (GTK_TREE_VIEW_COLUMN (column), 100);
@@ -418,16 +480,22 @@ row_activate_cb(GtkTreeView *tree_view,
 }
 
 static void
-set_convert_result(struct app_win *win, const char *str, const char *wt)
+set_convert_result(struct app_win *win, struct words *w,
+		   const char *index_str,
+		   const char *word_str, const char *wt)
 {
     char buf[1024];
+    char buf2[1024];
     struct anthy_segment_stat ss;
     int i;
     char *tail = syusi_tail(wt);
-    sprintf(buf, "%s%s", str, tail);
+    sprintf(buf, "%s%s", index_str, tail);
     printf("%s\n", buf);
     anthy_set_string(win->ac, buf);
     anthy_get_segment_stat(win->ac, 0, &ss);
+    /**/
+    sprintf(buf2, "%s%s", word_str, tail);
+    /**/
     gtk_list_store_clear(GTK_LIST_STORE(win->res_model));
     for (i = 0; i < ss.nr_candidate; i++) {
 	GtkTreeIter iter;
@@ -438,6 +506,16 @@ set_convert_result(struct app_win *win, const char *str, const char *wt)
 	gtk_list_store_set(GTK_LIST_STORE(win->res_model), &iter,
 			   0, win->cands[i],
 			   -1);
+	/**/
+	if (i == 0) {
+	    if (!strcmp(buf2, buf)) {
+		w->is_top = 1;
+		printf("top\n");
+	    } else {
+		w->is_top = 0;
+		printf("not top\n");
+	    }
+	}
     }
 }
 
@@ -458,7 +536,7 @@ selection_cb(GtkTreeSelection *selection,
     w = g_value_get_pointer(&value);
     g_value_unset(&value);
     if (w) {
-	set_convert_result(win, w->index, w->wt);
+	set_convert_result(win, w, w->index, w->word, w->wt);
     }
 }
 
@@ -484,9 +562,11 @@ build_tree_view(struct app_win *win, GtkWidget *hbox)
 
     gtk_widget_show(sw);
 
-    g_signal_connect(win->words_view, "row_activated", row_activate_cb, NULL);
+    g_signal_connect(win->words_view, "row_activated",
+		     G_CALLBACK(row_activate_cb), NULL);
     selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(win->words_view));
-    g_signal_connect(selection, "changed", selection_cb, win);
+    g_signal_connect(selection, "changed",
+		     G_CALLBACK(selection_cb), win);
 }
 
 static void
@@ -507,7 +587,7 @@ build_result_view(struct app_win *win)
     GtkTreeViewColumn *column;
 
     GtkWidget *label = gtk_label_new("result");
-    win->res_model = gtk_list_store_new(1, G_TYPE_STRING);
+    win->res_model = GTK_TREE_MODEL(gtk_list_store_new(1, G_TYPE_STRING));
     win->res_view = gtk_tree_view_new_with_model(win->res_model);
     renderer = gtk_cell_renderer_text_new();
     column = gtk_tree_view_column_new_with_attributes("word",
@@ -568,6 +648,36 @@ build_ui(struct app_win *win)
     gtk_widget_show(win->top);
 }
 
+static void
+load_checked(struct app_win *win)
+{
+    FILE *fp;
+    char buf[1024];
+    int nr = 0;
+    fp = fopen(CHECKED_ITEM_FILE, "r");
+    if (!fp) {
+	printf("no checked\n");
+	return ;
+    }
+    while (fgets(buf, 1024, fp)) {
+	char index[1024];
+	char word[1024];
+	char wt[1024];
+	char fix[1024];
+	struct words *w;
+	if (sscanf(buf, "%s %s %s %s", index, word, wt, fix) != 4) {
+	    continue;
+	}
+	w = find_word(&win->dict, index, word, wt);
+	if (w) {
+	    nr++;
+	    w->checked = 1;
+	}
+    }
+    fclose(fp);
+    printf("%d checked\n", nr);
+}
+
 static void
 init_anthy(struct app_win *win)
 {
@@ -585,6 +695,7 @@ start_cb(GtkWidget *button, struct app_win *win)
 {
     build_ui(win);
     load_words(win);
+    load_checked(win);
     update_model(win);
     gtk_widget_hide(win->conf_win);
 }
@@ -602,7 +713,8 @@ build_conf_win(struct app_win *win)
     gtk_widget_show(button);
     gtk_widget_show(vbox);
     gtk_widget_show(win->conf_win);
-    gtk_signal_connect(button, "clicked", start_cb, win);
+    g_signal_connect(button, "clicked",
+		     G_CALLBACK(start_cb), win);
 }
 
 int
